# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Integration tests for 4C locsys conditions."""

import pytest

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.function import Function
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.locsys_condition import LocSysCondition
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_integration_four_c_locsys_condition_locsys(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test case for point locsys condition for beams.

    The testcase is adapted from to
    beam3r_herm2line3_static_locsys.4C.yaml. However it has a simpler
    material, and an additional line locsys condition.
    """

    locsys_rotation = Rotation([0, 0, 1], 0.1)

    # Create the mesh.
    mesh = Mesh()

    fun = Function([{"SYMBOLIC_FUNCTION_OF_SPACE_TIME": "t"}])
    mesh.add(fun)

    # Create the beam.
    beam_set = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        get_default_test_beam_material(material_type="reissner"),
        [2.5, 2.5, 2.5],
        [4.5, 2.5, 2.5],
        n_el=1,
    )

    # Add dirichlet boundary conditions.
    mesh.add(
        BoundaryCondition(
            beam_set["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    # Add additional dirichlet boundary condition to check if combination with locsys condition works.
    mesh.add(
        BoundaryCondition(
            beam_set["end"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 0, 0, 0, 0, 0, 0, 0, 0],
                "VAL": [1.0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [fun, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )

    # Add locsys condition with rotation
    mesh.add(LocSysCondition(beam_set["end"], rotation=locsys_rotation))

    # Add line Function with function array

    fun_2 = Function([{"SYMBOLIC_FUNCTION_OF_SPACE_TIME": "2.0*t"}])
    mesh.add(fun_2)

    # Check if the LocSys condition is added correctly for a line with additional options.
    mesh.add(
        LocSysCondition(
            GeometrySet(beam_set["line"]),
            rotation=locsys_rotation,
            function_array=[fun_2],
            update_node_position=True,
            use_consistent_node_normal=True,
        )
    )

    # Check that the Locsys condition works with 3 given functions
    mesh.add(
        LocSysCondition(
            GeometrySet(beam_set["start"]),
            function_array=[fun, fun_2, fun_2],
            update_node_position=True,
            use_consistent_node_normal=False,
        )
    )

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), mesh)

    # Check that the combination of 3 functions and a rotation raises an error
    with pytest.raises(ValueError, match="If more than a single"):
        mesh.add(
            LocSysCondition(
                GeometrySet(beam_set["start"]),
                rotation=locsys_rotation,
                function_array=[fun, fun_2, fun_2],
                update_node_position=True,
                use_consistent_node_normal=False,
            )
        )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Functions to create meshed beam geometries at a node."""

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.mesh_creation_functions.beam_arc import (
    create_beam_mesh_arc_segment_via_axis as _create_beam_mesh_arc_segment_via_axis,
)
from beamme.mesh_creation_functions.beam_line import (
    create_beam_mesh_line as _create_beam_mesh_line,
)
from beamme.utils.nodes import get_single_node as _get_single_node


def create_beam_mesh_line_at_node(
    mesh, beam_class, material, start_node, length, **kwargs
):
    """Generate a straight line at a given node. The tangent will be the same
    as at that node.

    Args
    ----
    mesh: Mesh
        Mesh that the arc segment will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this segment.
    start_node: _np.array, list
        Point where the arc will continue.
    length: float
        Length of the line.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    if length < 0:
        raise ValueError("Length has to be positive!")

    # Create the line starting from the given node
    start_node = _get_single_node(start_node)
    tangent = start_node.rotation * [1, 0, 0]
    start_position = start_node.coordinates
    end_position = start_position + tangent * length

    return _create_beam_mesh_line(
        mesh,
        beam_class,
        material,
        start_position,
        end_position,
        start_node=start_node,
        **kwargs,
    )


def create_beam_mesh_arc_at_node(
    mesh, beam_class, material, start_node, arc_axis_normal, radius, angle, **kwargs
):
    """Generate a circular segment starting at a given node. The arc will be
    tangent to the given node.

    Args
    ----
    mesh: Mesh
        Mesh that the arc segment will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this segment.
    start_node: _np.array, list
        Point where the arc will continue.
    arc_axis_normal: 3d-vector
        Rotation axis for the created arc.
    radius: float
        The radius of the arc segment.
    angle: float
        Angle of the arc. If the angle is negative, the arc will point in the
        opposite direction, i.e., as if the arc_axis_normal would change sign.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    # If the angle is negative, the normal is switched
    arc_axis_normal = _np.asarray(arc_axis_normal)
    if angle < 0:
        arc_axis_normal = -1.0 * arc_axis_normal

    # The normal has to be perpendicular to the start point tangent
    start_node = _get_single_node(start_node)
    tangent = start_node.rotation * [1, 0, 0]
    if _np.abs(_np.dot(tangent, arc_axis_normal)) > _bme.eps_pos:
        raise ValueError(
            "The normal has to be perpendicular to the tangent in the start node!"
        )

    # Get the center of the arc
    center_direction = _np.cross(tangent, arc_axis_normal)
    center_direction *= 1.0 / _np.linalg.norm(center_direction)
    center = start_node.coordinates - center_direction * radius

    return _create_beam_mesh_arc_segment_via_axis(
        mesh,
        beam_class,
        material,
        arc_axis_normal,
        center,
        start_node.coordinates,
        _np.abs(angle),
        start_node=start_node,
        **kwargs,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file defines the base beam element."""

from typing import Any as _Any
from typing import Callable as _Callable
from typing import List as _List
from typing import Optional as _Optional
from typing import Type as _Type

import numpy as _np
import vtk as _vtk

from beamme.core.conf import bme as _bme
from beamme.core.element import Element as _Element
from beamme.core.material import MaterialBeamBase as _MaterialBeamBase
from beamme.core.node import NodeCosserat as _NodeCosserat
from beamme.core.rotation import Rotation as _Rotation
from beamme.core.vtk_writer import add_point_data_node_sets as _add_point_data_node_sets


class Beam(_Element):
    """A base class for a beam element."""

    # An array that defines the parameter positions of the element nodes,
    # in ascending order.
    nodes_create: _Any = []

    # A list of valid material types for this element.
    valid_materials: _Any = []

    def __init__(self, material=None, nodes=None):
        super().__init__(nodes=nodes, material=material)

    @classmethod
    def create_beam(
        cls: _Type["Beam"],
        material: _MaterialBeamBase,
        beam_function: _Callable,
        *,
        start_node: _Optional[_NodeCosserat] = None,
        end_node: _Optional[_NodeCosserat] = None,
        relative_twist: _Optional[_Rotation] = None,
        set_nodal_arc_length: bool = False,
        nodal_arc_length_offset: _Optional[float] = None,
    ) -> tuple["Beam", _List[_NodeCosserat]]:
        """Create the nodes for this beam element. The function returns a list
        with the created nodes.

        In the case of start_node and end_node, it is checked, that the
        function and the node have the same coordinates and rotations.

        Args:
            material: The material to be used with this beam.
            beam_function:
                Returns the position, rotation and (optionally) arc length of the
                beam along the local coordinate xi. If no arc lengths is provided,
                the returned value should simply be None.
            start_node:
                If this argument is given, this is the node of the beam at xi=-1.
            end_node:
                If this argument is given, this is the node of the beam at xi=1.
            relative_twist:
                Apply this relative rotation to all created nodes. This can be used to
                "twist" the created beam to match the rotation of given nodes.
            set_nodal_arc_length:
                Flag if the arc length in the created nodes should be set.
            nodal_arc_length_offset:
                Offset of the stored nodal arc length w.r.t. to the one generated
                by the function.

        Returns:
            The created beam object and the newly created nodes. The created nodes
            are returned, because we can then directly add them to a `Mesh` without
            having to check if they are already in the mesh.
        """

        def check_node(node, pos, rot, arc_length, name):
            """Check if the given node matches with the position and rotation
            and optionally also the arc length."""

            if _np.linalg.norm(pos - node.coordinates) > _bme.eps_pos:
                raise ValueError(
                    f"{name} position does not match with function! Got {pos} from function but "
                    + f"given node value is {node.coordinates}"
                )
            if not node.rotation == rot:
                raise ValueError(f"{name} rotation does not match with function!")

            if arc_length is not None:
                if _np.abs(node.arc_length - arc_length) > _bme.eps_pos:
                    raise ValueError(
                        f"Arc lengths don't match, got {node.arc_length} and {arc_length}"
                    )

        # Flags if nodes are given
        has_start_node = start_node is not None
        has_end_node = end_node is not None

        # Create the element
        beam = cls(material=material)

        # Loop over local nodes.
        arc_length = None
        for i, xi in enumerate(cls.nodes_create):
            # Get the position and rotation at xi
            pos, rot, arc_length_from_function = beam_function(xi)
            if relative_twist is not None:
                rot = rot * relative_twist
            if set_nodal_arc_length:
                arc_length = arc_length_from_function + nodal_arc_length_offset

            # Check if the position and rotation match existing nodes
            if i == 0 and has_start_node:
                check_node(start_node, pos, rot, arc_length, "start_node")
                beam.nodes = [start_node]
            elif (i == len(cls.nodes_create) - 1) and has_end_node:
                check_node(end_node, pos, rot, arc_length, "end_node")

            # Create the node
            if (i > 0 or not has_start_node) and (
                i < len(cls.nodes_create) - 1 or not has_end_node
            ):
                is_middle_node = 0 < i < len(cls.nodes_create) - 1
                beam.nodes.append(
                    _NodeCosserat(
                        pos, rot, is_middle_node=is_middle_node, arc_length=arc_length
                    )
                )

        # Get a list with the created nodes.
        if has_start_node:
            created_nodes = beam.nodes[1:]
        else:
            created_nodes = beam.nodes.copy()

        # Add the end node to the beam.
        if has_end_node:
            beam.nodes.append(end_node)

        # Return the created beam object and the created nodes.
        return beam, created_nodes

    @classmethod
    def get_coupling_dict(cls, coupling_dof_type):
        """Return the dict to couple this beam to another beam."""

        match coupling_dof_type:
            case _bme.coupling_dof.joint:
                if cls.coupling_joint_dict is None:
                    raise ValueError(f"Joint coupling is not implemented for {cls}")
                return cls.coupling_joint_dict
            case _bme.coupling_dof.fix:
                if cls.coupling_fix_dict is None:
                    raise ValueError("Fix coupling is not implemented for {cls}")
                return cls.coupling_fix_dict
            case _:
                raise ValueError(
                    f'Coupling_dof_type "{coupling_dof_type}" is not implemented!'
                )

    def flip(self):
        """Reverse the nodes of this element.

        This is usually used when reflected.
        """
        self.nodes = [self.nodes[-1 - i] for i in range(len(self.nodes))]

    def _check_material(self):
        """Check if the linked material is valid for this type of beam
        element."""
        for material_type in type(self).valid_materials:
            if isinstance(self.material, material_type):
                break
        else:
            raise TypeError(
                f"Beam of type {type(self)} can not have a material of type {type(self.material)}!"
            )

    def get_vtk(
        self,
        vtk_writer_beam,
        vtk_writer_solid,
        *,
        beam_centerline_visualization_segments=1,
        **kwargs,
    ):
        """Add the representation of this element to the VTK writer as a poly
        line.

        Args
        ----
        vtk_writer_beam:
            VTK writer for the beams.
        vtk_writer_solid:
            VTK writer for solid elements, not used in this method.
        beam_centerline_visualization_segments: int
            Number of segments to be used for visualization of beam centerline between successive
            nodes. Default is 1, which means a straight line is drawn between the beam nodes. For
            Values greater than 1, a Hermite interpolation of the centerline is assumed for
            visualization purposes.
        """

        n_nodes = len(self.nodes)
        n_segments = n_nodes - 1
        n_additional_points_per_segment = beam_centerline_visualization_segments - 1
        # Number of points (in addition to the nodes) to be used for output
        n_additional_points = n_segments * n_additional_points_per_segment
        n_points = n_nodes + n_additional_points

        # Dictionary with cell data.
        cell_data = self.vtk_cell_data.copy()
        if self.material.radius is not None:
            cell_data["cross_section_radius"] = self.material.radius

        # Dictionary with point data.
        point_data = {}
        point_data["node_value"] = _np.zeros(n_points)
        point_data["base_vector_1"] = _np.zeros((n_points, 3))
        point_data["base_vector_2"] = _np.zeros((n_points, 3))
        point_data["base_vector_3"] = _np.zeros((n_points, 3))

        coordinates = _np.zeros((n_points, 3))
        nodal_rotation_matrices = [
            node.rotation.get_rotation_matrix() for node in self.nodes
        ]

        for i_node, (node, rotation_matrix) in enumerate(
            zip(self.nodes, nodal_rotation_matrices)
        ):
            coordinates[i_node, :] = node.coordinates
            if node.is_middle_node:
                point_data["node_value"][i_node] = 0.5
            else:
                point_data["node_value"][i_node] = 1.0

            point_data["base_vector_1"][i_node] = rotation_matrix[:, 0]
            point_data["base_vector_2"][i_node] = rotation_matrix[:, 1]
            point_data["base_vector_3"][i_node] = rotation_matrix[:, 2]

        # We can output the element partner index, which is a debug quantity to help find elements
        # with matching middle nodes. This is usually an indicator for an issue with the mesh.
        # TODO: Check if we really need this partner index output any more
        element_partner_indices = list(
            [
                node.element_partner_index
                for node in self.nodes
                if node.element_partner_index is not None
            ]
        )
        if len(element_partner_indices) > 1:
            raise ValueError(
                "More than one element partner indices are currently not supported in the output"
                "functionality"
            )
        elif len(element_partner_indices) == 1:
            cell_data["partner_index"] = element_partner_indices[0] + 1

        # Check if we have everything we need to write output or if we need to calculate additional
        # points for a smooth beam visualization.
        if beam_centerline_visualization_segments == 1:
            point_connectivity = _np.arange(n_nodes)
        else:
            # We need the centerline shape function matrices, so calculate them once and use for
            # all segments that we need. Drop the first and last value, since they represent the
            # nodes which we have already added above.
            xi = _np.linspace(-1, 1, beam_centerline_visualization_segments + 1)[1:-1]
            hermite_shape_functions_pos = _np.array(
                [
                    0.25 * (2.0 + xi) * (1.0 - xi) ** 2,
                    0.25 * (2.0 - xi) * (1.0 + xi) ** 2,
                ]
            ).transpose()
            hermite_shape_functions_tan = _np.array(
                [
                    0.125 * (1.0 + xi) * (1.0 - xi) ** 2,
                    -0.125 * (1.0 - xi) * (1.0 + xi) ** 2,
                ]
            ).transpose()

            point_connectivity = _np.zeros(n_points, dtype=int)

            for i_segment in range(n_segments):
                positions = _np.array(
                    [
                        self.nodes[i_node].coordinates
                        for i_node in [i_segment, i_segment + 1]
                    ]
                )
                tangents = _np.array(
                    [
                        nodal_rotation_matrices[i_node][:, 0]
                        for i_node in [i_segment, i_segment + 1]
                    ]
                )
                length_factor = _np.linalg.norm(positions[1] - positions[0])
                interpolated_coordinates = _np.dot(
                    hermite_shape_functions_pos, positions
                ) + length_factor * _np.dot(hermite_shape_functions_tan, tangents)

                index_first_point = (
                    n_nodes + i_segment * n_additional_points_per_segment
                )
                index_last_point = (
                    n_nodes + (i_segment + 1) * n_additional_points_per_segment
                )

                coordinates[index_first_point:index_last_point] = (
                    interpolated_coordinates
                )
                point_connectivity[
                    i_segment * beam_centerline_visualization_segments
                ] = i_segment
                point_connectivity[
                    (i_segment + 1) * beam_centerline_visualization_segments
                ] = i_segment + 1
                point_connectivity[
                    i_segment * beam_centerline_visualization_segments + 1 : (
                        i_segment + 1
                    )
                    * beam_centerline_visualization_segments
                ] = _np.arange(index_first_point, index_last_point)

        # Get the point data sets and add everything to the output file.
        _add_point_data_node_sets(
            point_data, self.nodes, extra_points=n_additional_points
        )
        indices = vtk_writer_beam.add_points(coordinates, point_data=point_data)
        vtk_writer_beam.add_cell(
            _vtk.vtkPolyLine, indices[point_connectivity], cell_data=cell_data
        )


def generate_beam_class(n_nodes: int):
    """Return a class representing a general beam with n_nodes in BeamMe.

    Args:
        n_nodes: Number of equally spaced nodes along the beam centerline.

    Returns:
        A beam object that has n_nodes along the centerline.
    """

    # Define the class variable responsible for creating the nodes.
    nodes_create = _np.linspace(-1, 1, num=n_nodes)

    # Create the beam class which inherits from the base beam class.
    return type(f"Beam{n_nodes}", (Beam,), {"nodes_create": nodes_create})


Beam2 = generate_beam_class(2)
Beam3 = generate_beam_class(3)
Beam4 = generate_beam_class(4)
Beam5 = generate_beam_class(5)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the examples."""

import pytest
from testbook import testbook


@pytest.mark.parametrize(
    "notebook_path",
    [
        "examples/example_1_finite_rotations.ipynb",
        "examples/example_2_core_mesh_generation_functions.ipynb",
    ],
)
def test_other_examples_notebooks(notebook_path):
    """Parameterized test case for multiple Jupyter notebooks.

    The notebook is run and it is checked that it runs through without
    any errors/assertions.
    """

    with testbook(notebook_path) as tb:
        # we do not define the examples as modules, therefore we need to add the
        # examples folder to the current sys path so examples/utils can be imported
        # within the notebooks correctly
        tb.inject(
            """
            import sys
            import os
            sys.path.insert(0, os.path.abspath(os.path.join(os.getcwd(), "examples")))
            """
        )

        # execute the notebook
        tb.execute()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines a global object that manages all kind of stuff."""

from enum import Enum as _Enum
from enum import auto as _auto

# Lines to be added to each created input file
INPUT_FILE_HEADER = [
    "-" * 40,
    "This input file was created with BeamMe.",
    "Copyright (c) 2018-2026 BeamMe Authors",
    "https://beamme-py.github.io/beamme/",
    "-" * 40,
]


class Geometry(_Enum):
    """Enum for geometry types."""

    point = _auto()
    line = _auto()
    surface = _auto()
    volume = _auto()


class BoundaryCondition(_Enum):
    """Enum for boundary condition types."""

    dirichlet = _auto()
    neumann = _auto()
    locsys = _auto()
    moment_euler_bernoulli = _auto()
    beam_to_beam_contact = _auto()
    beam_to_solid_volume_meshtying = _auto()
    beam_to_solid_surface_meshtying = _auto()
    beam_to_solid_surface_contact = _auto()
    point_coupling = _auto()
    point_coupling_penalty = _auto()
    point_coupling_indirect = _auto()

    def is_point_coupling_pairwise(self) -> bool:
        """Check whether the point coupling condition should be applied
        pairwise.

        Returns:
            bool: True if the coupling should be applied individually between pairs of nodes,
                rather than to the entire geometry set as a whole.
        """
        if self == self.point_coupling:
            return False
        elif self == self.point_coupling_penalty:
            return True
        else:
            raise TypeError(f"Got unexpected coupling type: {self}")


class CouplingDofType(_Enum):
    """Enum for coupling types."""

    fix = _auto()
    joint = _auto()


class DoubleNodes(_Enum):
    """Enum for handing double nodes in Neumann conditions."""

    remove = _auto()
    keep = _auto()


class BeamMe(object):
    """A global object that stores options for the whole BeamMe application."""

    def __init__(self):
        self.set_default_values()

        # Geometry types.
        self.geo = Geometry

        # Boundary conditions types.
        self.bc = BoundaryCondition

        # Coupling types.
        self.coupling_dof = CouplingDofType

        # Handling of multiple nodes in Neumann bcs.
        self.double_nodes = DoubleNodes

    def set_default_values(self):
        """Set the configuration to the default values."""

        # Set the epsilons for comparison of different types of values.
        self.eps_quaternion = 1e-10
        self.eps_pos = 1e-10
        self.eps_knot_vector = 1e-10

        # Allow the rotation of beams when connected and the triads do not
        # match.
        self.allow_beam_rotation = True

        # Check for overlapping elements when creating an input file.
        self.check_overlapping_elements = True


bme = BeamMe()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to unittest the functionality of the couplings."""

import numpy as np
import pytest

from beamme.core.conf import bme
from beamme.core.coupling import Coupling, coupling_factory
from beamme.core.geometry_set import GeometrySet
from beamme.core.node import Node, NodeCosserat
from beamme.core.rotation import Rotation


def test_beamme_core_coupling_factory():
    """Test that the coupling factory can be used with the desired inputs."""

    nodes = [Node([0, 0, 0]) for i in range(4)]
    node_set = GeometrySet(nodes)

    # Single coupling for all nodes, list of nodes given
    couplings = coupling_factory(nodes, bme.bc.point_coupling, {})
    assert len(couplings) == 1
    coupling_points = couplings[0].geometry_set.get_points()
    assert len(coupling_points) == 4
    for i in range(4):
        assert coupling_points[i] is nodes[i]

    # Single coupling for all nodes, geometry set given
    couplings = coupling_factory(node_set, bme.bc.point_coupling, {})
    assert len(couplings) == 1
    assert couplings[0].geometry_set is node_set

    # Pairwise couplings, list of nodes given
    couplings = coupling_factory(nodes, bme.bc.point_coupling_penalty, {})
    assert len(couplings) == 3
    for i_coupling in range(3):
        coupling_nodes = couplings[i_coupling].geometry_set.get_points()
        assert len(coupling_nodes) == 2
        assert coupling_nodes[0] is nodes[0]
        assert coupling_nodes[1] is nodes[i_coupling + 1]

    # Pairwise couplings, geometry set given
    couplings = coupling_factory(node_set, bme.bc.point_coupling_penalty, {})
    assert len(couplings) == 3
    reference_geometry_set_nodes = node_set.get_points()
    for i_coupling in range(3):
        coupling_nodes = couplings[i_coupling].geometry_set.get_points()
        assert len(coupling_nodes) == 2
        assert coupling_nodes[0] is reference_geometry_set_nodes[0]
        assert coupling_nodes[1] is reference_geometry_set_nodes[i_coupling + 1]


@pytest.mark.parametrize("within_tolerance", (True, False))
@pytest.mark.parametrize("check_overlapping_nodes", (True, False))
def test_beamme_core_coupling_check_overlapping_coupling_nodes(
    within_tolerance, check_overlapping_nodes
):
    """Per default, we check that coupling nodes are at the same physical
    position.

    This check can be deactivated with the keyword
    check_overlapping_nodes when creating a Coupling.
    """

    # Create the nodes
    factor = 0.5 if within_tolerance else 2.0
    ref_position = [0.0, 1.5, 3.0]
    node_1 = NodeCosserat(ref_position, Rotation())
    node_2 = NodeCosserat(
        ref_position + factor * bme.eps_pos * np.array([1.0, 1.0, 1.0]), Rotation()
    )

    # Couple two nodes whose positions may be within or outside the positional
    # tolerance. When check_overlapping_nodes is enabled and the nodes are
    # outside the tolerance, creating the coupling is expected to raise an error.
    args = [[node_1, node_2], bme.bc.point_coupling, "coupling_type_string"]
    if check_overlapping_nodes and not within_tolerance:
        with pytest.raises(ValueError):
            Coupling(*args)
    else:
        Coupling(*args, check_overlapping_nodes=False)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements the class that represents one element in the Mesh."""

from beamme.core.base_mesh_item import BaseMeshItem as _BaseMeshItem


class Element(_BaseMeshItem):
    """A base class for an FEM element in the mesh."""

    def __init__(self, nodes=None, material=None, **kwargs):
        super().__init__(**kwargs)

        # List of nodes that are connected to the element.
        if nodes is None:
            self.nodes = []
        else:
            self.nodes = nodes.copy()

        # Material of this element.
        self.material = material

        # VTK cell data for this element.
        self.vtk_cell_data = {}

    def flip(self):
        """Reverse the nodes of this element.

        This is usually used when reflected.
        """
        raise NotImplementedError(
            f"The flip method is not implemented for {self.__class__}"
        )

    def replace_node(self, old_node, new_node):
        """Replace old_node with new_node."""

        # Look for old_node and replace it. If it is not found, throw error.
        for i, node in enumerate(self.nodes):
            if node == old_node:
                self.nodes[i] = new_node
                break
        else:
            raise ValueError(
                "The node that should be replaced is not in the current element"
            )

    def get_vtk(self, vtk_writer_beam, vtk_writer_solid, **kwargs):
        """Add representation of this element to the vtk_writers for solid and
        beam."""
        raise NotImplementedError("VTK output has to be implemented in the class!")



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file defines the interface to the Scipy spatial geometric search
functionality."""

from scipy.spatial import KDTree as _KDTree


def pairs_to_partner_list(pairs, n_points):
    """Convert the pairs to a partner list."""

    # Sort the pairs by the first column
    pairs = pairs[pairs[:, 0].argsort()]

    n_partners = 0
    partner_index_list = [-1 for i in range(n_points)]
    for pair in pairs:
        pair_partners = [partner_index_list[pair[i]] for i in range(2)]
        if pair_partners[0] == -1 and pair_partners[1] == -1:
            # We have a new partner index
            for i in range(2):
                partner_index_list[pair[i]] = n_partners
            n_partners += 1
        elif pair_partners[0] == -1:
            partner_index_list[pair[0]] = pair_partners[1]
        elif pair_partners[1] == -1:
            partner_index_list[pair[1]] = pair_partners[0]
        else:
            # Both points already have a partner index. Since we order
            # the pairs this one has to be equal to each other or we
            # have multiple clusters.
            if not pair_partners[0] == pair_partners[1]:
                raise ValueError(
                    "Two points are connected to different partner indices."
                )
    return partner_index_list, n_partners


def find_close_points_scipy(point_coordinates, tol):
    """Call the Scipy implementation of find close_points."""

    kd_tree = _KDTree(point_coordinates)
    pairs = kd_tree.query_pairs(r=tol, output_type="ndarray")
    return pairs_to_partner_list(pairs, len(point_coordinates))



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file has functions to generate multiple parallel fibers within a
rectangle.

This can for example be used to create fiber reinforced composite
plates.
"""

import numpy as _np

from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.mesh_creation_functions.beam_line import (
    create_beam_mesh_line as _create_beam_mesh_line,
)
from beamme.utils.nodes import check_node_by_coordinate as _check_node_by_coordinate


def _intersect_line_with_rectangle(
    length, width, start_line, direction_line, fail_if_no_intersection=True
):
    """Calculate the intersection points between a line and a rectangle.

    Args
    ----
    length: scalar
        Rectangle length in x direction.
    width: scalar
        Rectangle width in y direction.
    start_line: 2d-list
        Point on the line.
    direction_line: 2d-list
        Direction of the line.
    fail_if_no_intersection: bool
        If this is true and no intersections are found, an error will be
        thrown.

    Return
    ----
    (start, end, projection_found)
    start: 2D vector
        Start point of intersected line.
    end: 2D vector
        End point of intersected line.
    projection_found: bool
        True if intersection is valid.
    """

    # Convert the input values to np.arrays.
    start_line = _np.asarray(start_line)
    direction_line = _np.asarray(direction_line)

    # Set definition for the boundary lines of the rectangle. The director is
    # chosen in a way, that the values [0, 1] for the line parameters alpha are
    # valid.
    # boundary_lines = [..., [start, dir], ...]
    boundary_lines = [
        [[0, 0], [length, 0]],
        [[0, 0], [0, width]],
        [[0, width], [length, 0]],
        [[length, 0], [0, width]],
    ]
    # Convert to numpy arrays.
    boundary_lines = [
        [_np.array(item) for item in boundary] for boundary in boundary_lines
    ]

    # Loop over the boundaries.
    alpha_list = []
    for start_boundary, direction_boundary in boundary_lines:
        # Set up the linear system to solve the intersection problem.
        A = _np.transpose(_np.array([direction_line, -direction_boundary]))

        # Check if the system is solvable.
        if _np.abs(_np.linalg.det(A)) > 1e-10:
            alpha = _np.linalg.solve(A, start_boundary - start_line)
            if 0 <= alpha[1] and alpha[1] <= 1:
                alpha_list.append(alpha[0])

    # Check that intersections were found.
    if len(alpha_list) < 2:
        if fail_if_no_intersection:
            raise ValueError("No intersections found!")
        return (None, None, False)

    # Return the start and end point on the line.
    return (
        start_line + min(alpha_list) * direction_line,
        start_line + max(alpha_list) * direction_line,
        True,
    )


def create_fibers_in_rectangle(
    mesh,
    beam_class,
    material,
    length,
    width,
    angle,
    fiber_normal_distance,
    fiber_element_length,
    *,
    reference_point=None,
    fiber_element_length_min=None,
):
    """Create multiple fibers in a rectangle.

    Args
    ----
    mesh: Mesh
        Mesh that the fibers will be added to.
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    length: float
        Length of the rectangle in x direction (starting at x=0)
    width: float
        Width of the rectangle in y direction (starting at y=0)
    angle: float
        Angle of the fibers in degree.
    fiber_normal_distance: float
        Normal distance between the parallel fibers.
    fiber_element_length: float
        Length of a single beam element. In general it will not be possible to
        exactly achieve this length.
    reference_point: [float, float]
        Specify a single point inside the rectangle that one of the fibers will pass through.
        Per default the reference point is in the middle of the rectangle.
    fiber_element_length_min: float
        Minimum fiber length. If a fiber is shorter than this value, it will not be created.
        The default value is half of fiber_element_length.
    """

    if reference_point is None:
        reference_point = 0.5 * _np.array([length, width])
    else:
        if (
            reference_point[0] < 0.0
            or reference_point[0] > length
            or reference_point[1] < 0.0
            or reference_point[1] > width
        ):
            raise ValueError("The reference point has to lie within the rectangle")

    if fiber_element_length_min is None:
        fiber_element_length_min = 0.5 * fiber_element_length
    elif fiber_element_length_min < 0.0:
        raise ValueError("fiber_element_length_min must be positive!")

    # Get the fiber angle in rad.
    fiber_angle = angle * _np.pi / 180.0
    sin = _np.sin(fiber_angle)
    cos = _np.cos(fiber_angle)

    # Direction and normal vector of the fibers.
    fiber_direction = _np.array([cos, sin])
    fiber_normal = _np.array([-sin, cos])

    # Get an upper bound of the number of fibers in this layer.
    diagonal = _np.sqrt(length**2 + width**2)
    fiber_n_max = int(_np.ceil(diagonal / fiber_normal_distance)) + 1

    # Go in both directions from the start point.
    for direction_sign, n_start in [[-1, 1], [1, 0]]:
        # Create a fiber as long as an intersection is found.
        for i_fiber in range(n_start, fiber_n_max):
            # Get the start and end point of the line.
            start, end, projection_found = _intersect_line_with_rectangle(
                length,
                width,
                reference_point
                + fiber_normal * i_fiber * fiber_normal_distance * direction_sign,
                fiber_direction,
                fail_if_no_intersection=False,
            )

            if projection_found:
                # Calculate the length of the line.
                fiber_length = _np.linalg.norm(end - start)

                # Create the beams if the length is not smaller than the fiber
                # distance.
                if fiber_length >= fiber_element_length_min:
                    # Calculate the number of elements in this fiber.
                    fiber_nel = int(_np.round(fiber_length / fiber_element_length))
                    fiber_nel = _np.max([fiber_nel, 1])
                    _create_beam_mesh_line(
                        mesh,
                        beam_class,
                        material,
                        _np.append(start, 0.0),
                        _np.append(end, 0.0),
                        n_el=fiber_nel,
                    )
            else:
                # The current search position is already outside of the rectangle, no need to continue.
                break

    return_set = _GeometryName()
    return_set["north"] = _GeometrySet(
        mesh.get_nodes_by_function(_check_node_by_coordinate, 1, width),
    )
    return_set["east"] = _GeometrySet(
        mesh.get_nodes_by_function(_check_node_by_coordinate, 0, length),
    )
    return_set["south"] = _GeometrySet(
        mesh.get_nodes_by_function(_check_node_by_coordinate, 1, 0)
    )
    return_set["west"] = _GeometrySet(
        mesh.get_nodes_by_function(_check_node_by_coordinate, 0, 0)
    )
    return_set["all"] = _GeometrySet(mesh.elements)
    return return_set



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file provides functions to create Abaqus beam element classes."""

from beamme.core.element_beam import Beam2 as _Beam2
from beamme.core.element_beam import Beam3 as _Beam3


def generate_abaqus_beam(beam_type: str):
    """Return a class representing a beam in Abaqus. This class can be used in
    the standard mesh generation functions.

    Args:
        beam_type: Abaqus identifier for this beam element. For more details,
            have a look at the Abaqus manual on "Choosing a beam element".

    Returns:
        A class representing the Abaqus beam element. The class inherits from
        the BeamX class, depending on the number of nodes.
    """

    if not beam_type[0].lower() == "b":
        raise TypeError("Could not identify the given Abaqus beam element")

    n_dim = int(beam_type[1])
    element_type = int(beam_type[2])

    if not n_dim == 3:
        raise ValueError("Currently only 3D beams in Abaqus are supported")
    if element_type == 1:
        base_class = _Beam2
    elif element_type == 2:
        base_class = _Beam3
    elif element_type == 3:
        base_class = _Beam2
    else:
        raise ValueError(f"Got unexpected element_type {element_type}")

    # Create the Abaqus beam class.
    return type(
        "BeamAbaqus" + beam_type,
        (base_class,),
        {
            "beam_type": beam_type,
            "n_dim": n_dim,
        },
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test that the input files created for testing are up to date."""

import pytest

from tests.create_test_models import (
    create_beam_to_solid_conditions_model,
)


@pytest.mark.parametrize("full_import", (False, True))
def test_other_create_input_files_beam_to_solid(
    full_import,
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create the input file for some beam-to-solid tests."""

    # Create the input file
    input_file, mesh = create_beam_to_solid_conditions_model(
        get_default_test_beam_material,
        get_corresponding_reference_file_path,
        full_import=full_import,
    )
    input_file.add(mesh)

    # Check results
    assert_results_close(get_corresponding_reference_file_path(), input_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script validates all 4C input files with FourCIPP in this
repository."""

import sys as _sys
from dataclasses import dataclass as _dataclass
from pathlib import Path as _Path

from fourcipp.fourc_input import FourCInput


@_dataclass
class Error:
    """Class to store import errors."""

    filename: str
    message: str


def validate_file(filename: str) -> list[Error]:
    """Validate a 4C input file with FourCIPP.

    Args:
        filename: The filename of the 4C input file to validate.

    Returns:
        List of errors found in the 4C input file or None.
    """

    try:
        inputfile = FourCInput.from_4C_yaml(_Path(filename))
        inputfile.validate(sections_only=True)
    except Exception as e:
        return [Error(filename, str(e))]

    return []


def main() -> None:
    """Validate all 4C input files with FourCIPP."""

    errors: list[Error] = []

    for filename in _sys.argv[1:]:
        if not _Path(filename).exists():
            errors.append(Error(filename, "File not found!"))

        errors.extend(validate_file(filename))

    if errors:
        print("Found 4C input files with errors:\n")

        for error in errors:
            print(f"{error.filename}: {error.message}\n")

        print("Please ensure that all 4C input files are correctly validated.")
        _sys.exit(1)
    else:
        _sys.exit(0)


if __name__ == "__main__":
    """Run the main function if the script is executed."""
    main()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests beam elements for 4C.

TODO: Split these tests into actual unit tests for beam and material.
"""

import warnings

import numpy as np
import pytest

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.function import Function
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import (
    Beam3eb,
    get_four_c_kirchhoff_beam,
    get_four_c_reissner_beam,
)
from beamme.four_c.four_c_types import (
    BeamKirchhoffConstraintType,
    BeamKirchhoffParametrizationType,
)
from beamme.four_c.input_file import InputFile
from beamme.four_c.material import (
    MaterialEulerBernoulli,
    MaterialKirchhoff,
    MaterialReissner,
)
from beamme.mesh_creation_functions.beam_arc import (
    create_beam_mesh_arc_segment_via_rotation,
)
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.utils.nodes import get_single_node


@pytest.mark.parametrize(
    ("n_nodes", "is_hermite"),
    (
        (2, False),
        (3, False),
        (4, False),
        (5, False),
        (3, True),
    ),
)
def test_integration_four_c_element_beam_reissner_beam(
    n_nodes, is_hermite, assert_results_close, get_corresponding_reference_file_path
):
    """Test that the input file for all types of Reissner beams is generated
    correctly."""

    # Create mesh
    mesh = Mesh()

    # Create material
    material = MaterialReissner(
        radius=1.0, youngs_modulus=1.0, nu=0.3, density=1.0, interaction_radius=2.0
    )

    beam_type = get_four_c_reissner_beam(
        n_nodes=n_nodes, is_hermite_centerline=is_hermite
    )
    create_beam_mesh_arc_segment_via_rotation(
        mesh,
        beam_type,
        material,
        [0.0, 0.0, 0.0],
        Rotation([0.0, 0.0, 1.0], np.pi / 2.0),
        2.0,
        np.pi / 2.0,
        n_el=2,
    )

    # Compare with the reference solution.
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=(
                f"line{n_nodes}" + (f"_hermite" if is_hermite else "")
            ),
        ),
        mesh,
    )


def test_integration_four_c_element_beam_kirchhoff_beam(
    assert_results_close, get_corresponding_reference_file_path
):
    """Test that the input file for all types of Kirchhoff beams is generated
    correctly."""

    # Create mesh
    mesh = Mesh()

    with warnings.catch_warnings():
        # Ignore the warnings for the rotvec beams.
        warnings.simplefilter("ignore")

        # Loop over options.
        for is_fad in (True, False):
            material = MaterialKirchhoff(
                radius=1.0, youngs_modulus=1.0, nu=1.0, density=1.0, is_fad=is_fad
            )
            for constraint in BeamKirchhoffConstraintType:
                for parametrization in BeamKirchhoffParametrizationType:
                    # Define the beam object factory function for the
                    # creation functions.
                    BeamObject = get_four_c_kirchhoff_beam(
                        constraint=constraint,
                        parametrization=parametrization,
                        is_fad=is_fad,
                    )

                    # Create a beam.
                    set_1 = create_beam_mesh_line(
                        mesh,
                        BeamObject,
                        material,
                        [0, 0, 0],
                        [1, 0, 0],
                        n_el=2,
                    )
                    set_2 = create_beam_mesh_line(
                        mesh,
                        BeamObject,
                        material,
                        [1, 0, 0],
                        [2, 0, 0],
                        n_el=2,
                    )

                    # Couple the nodes.
                    if parametrization == BeamKirchhoffParametrizationType.rot:
                        mesh.couple_nodes(
                            nodes=[
                                get_single_node(set_1["end"]),
                                get_single_node(set_2["start"]),
                            ]
                        )

                    # Move the mesh away from the next created beam.
                    mesh.translate([0, 0.5, 0])

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_four_c_element_beam_euler_bernoulli(
    assert_results_close, get_corresponding_reference_file_path
):
    """Recreate the 4C test case beam3eb_static_endmoment_quartercircle.4C.yaml
    This tests the implementation for Euler Bernoulli beams."""

    # Create the mesh and add function and material.
    mesh = Mesh()
    fun = Function([{"COMPONENT": 0, "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "t"}])
    mesh.add(fun)
    mat = MaterialEulerBernoulli(radius=1.0, youngs_modulus=1.0, nu=0.3, density=1.0)

    # Create the beam.
    beam_set = create_beam_mesh_line(mesh, Beam3eb, mat, [-1, 0, 0], [1, 0, 0], n_el=16)

    # Add boundary conditions.
    mesh.add(
        BoundaryCondition(
            beam_set["start"],
            {
                "NUMDOF": 6,
                "ONOFF": [1, 1, 1, 0, 1, 1],
                "VAL": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                "FUNCT": [0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    mesh.add(
        BoundaryCondition(
            beam_set["end"],
            {
                "NUMDOF": 6,
                "ONOFF": [0, 0, 0, 0, 0, 1],
                "VAL": [0.0, 0.0, 0.0, 0.0, 0.0, 7.8539816339744e-05],
                "FUNCT": [0, 0, 0, 0, 0, fun],
            },
            bc_type=bme.bc.moment_euler_bernoulli,
        )
    )

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), mesh)

    # Test consistency checks.
    rot = Rotation([1, 2, 3], 2.3434)
    mesh.nodes[-1].rotation = rot
    with pytest.raises(
        ValueError,
        match="The two nodal rotations in Euler Bernoulli beams must be the same",
    ):
        # This raises an error because not all rotation in the beams are
        # the same.
        input_file = InputFile()
        input_file.add(mesh)

    for node in mesh.nodes:
        node.rotation = rot
    with pytest.raises(
        ValueError,
        match="The rotations do not match the direction of the Euler Bernoulli beam",
    ):
        # This raises an error because the rotations do not match the
        # director between the nodes.
        input_file = InputFile()
        input_file.add(mesh)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file has functions to create a beam from a parametric curve."""

import autograd.numpy as _npAD
import numpy as _np
import scipy.integrate as _integrate
import scipy.optimize as _optimize
from autograd import jacobian as _jacobian

from beamme.core.conf import bme as _bme
from beamme.core.rotation import Rotation as _Rotation
from beamme.core.rotation import smallest_rotation as _smallest_rotation
from beamme.mesh_creation_functions.beam_generic import (
    create_beam_mesh_generic as _create_beam_mesh_generic,
)


def create_beam_mesh_parametric_curve(
    mesh,
    beam_class,
    material,
    function,
    interval,
    *,
    output_length=False,
    function_derivative=None,
    function_rotation=None,
    **kwargs,
):
    """Generate a beam from a parametric curve. Integration along the beam is
    performed with scipy, and if the gradient is not explicitly provided, it is
    calculated with the numpy wrapper autograd.

    Args
    ----
    mesh: Mesh
        Mesh that the curve will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this line.
    function: function
        3D-parametric curve that represents the beam axis. If only a 2D
        point is returned, the triad creation is simplified. If
        mathematical functions are used, they have to come from the wrapper
        autograd.numpy.
    interval: [start end]
        Start and end values for the parameter of the curve.
    output_length: bool
        If this is true, the function returns a tuple containing the created
        sets and the total arc length along the integrated function.
    function_derivative: function -> R3
        Explicitly provide the jacobian of the centerline position.
    function_rotation: function -> Rotation
        If this argument is given, the triads are computed with this
        function, on the same interval as the position function. Must
        return a Rotation object.
        If no function_rotation is given, the rotation of the first node
        is calculated automatically and all subsequent nodal rotations
        are calculated based on a smallest rotation mapping onto the curve
        tangent vector.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the curve. Also a 'line' set
        with all nodes of the curve.
    """

    # Check size of position function
    if len(function(interval[0])) == 2:
        is_3d_curve = False
    elif len(function(interval[0])) == 3:
        is_3d_curve = True
    else:
        raise ValueError("Function must return either 2d or 3d curve!")

    # Check rotation function.
    if function_rotation is None:
        is_rot_funct = False
    else:
        is_rot_funct = True

    # Check that the position is an np.array
    if not isinstance(function(interval[0]), _np.ndarray):
        raise TypeError(
            "Function must be of type np.ndarray, got {}!".format(
                type(function(interval[0]))
            )
        )

    # Get the derivative of the position function and the increment along
    # the curve.
    if function_derivative is None:
        rp = _jacobian(function)
    else:
        rp = function_derivative

    # Check which one of the boundaries is larger.
    if interval[0] > interval[1]:
        # In this case rp needs to be negated.
        rp_positive = rp

        def rp(t):
            """Return the inverted tangent vector."""
            return -(rp_positive(t))

    def ds(t):
        """Increment along the curve."""
        return _npAD.linalg.norm(rp(t))

    def S(t, start_t=None, start_S=None):
        """Function that integrates the length until a parameter value.

        A speedup can be achieved by giving start_t and start_S, the
        parameter and Length at a known point.
        """
        if start_t is None and start_S is None:
            st = interval[0]
            sS = 0
        elif start_t is not None and start_S is not None:
            st = start_t
            sS = start_S
        else:
            raise ValueError("Input parameters are wrong!")
        return _integrate.quad(ds, st, t)[0] + sS

    def get_t_along_curve(arc_length, t0, **kwargs):
        """Calculate the parameter t where the length along the curve is
        arc_length.

        t0 is the start point for the Newton iteration.
        """
        t_root = _optimize.newton(lambda t: S(t, **kwargs) - arc_length, t0, fprime=ds)
        return t_root

    class BeamFunctions:
        """This object manages the creation of functions which are used to
        create the beam nodes.

        By wrapping this in this object, we can store some data and
        speed up the numerical integration along the curve.
        """

        def __init__(self):
            """Initialize the object."""
            self._reset_start_values()
            self.last_triad = None

            if is_3d_curve:
                r_prime = rp(interval[0])
                if abs(_np.dot(r_prime, [0, 0, 1])) < abs(_np.dot(r_prime, [0, 1, 0])):
                    t2_temp = [0, 0, 1]
                else:
                    t2_temp = [0, 1, 0]
                self.last_triad = _Rotation.from_basis(r_prime, t2_temp)

        def _reset_start_values(self):
            """Reset the stored start values for the next Newton iteration."""
            self.t_start_newton = interval[0]
            self.S_start_newton = 0.0

        def __call__(self, length_a, length_b):
            """This object is called with the interval limits.

            This method returns a function that evaluates the position
            and rotation within this interval.
            """

            # In case the interval is not continuous with the last one, we reset the start
            # values for the Newton iteration here
            if length_a < self.S_start_newton - _bme.eps_pos:
                self._reset_start_values()

            # Length of the beam element in physical space.
            L = length_b - length_a

            def get_beam_position_and_rotation_at_xi(xi):
                """Evaluate the beam position and rotation at xi.

                xi is the beam element parameter coordinate, i.e., xi =
                [-1, 1].
                """
                # Parameter for xi.
                S = length_a + 0.5 * (xi + 1) * L
                t = get_t_along_curve(
                    S,
                    self.t_start_newton,
                    start_t=self.t_start_newton,
                    start_S=self.S_start_newton,
                )

                # Position at xi.
                if is_3d_curve:
                    pos = function(t)
                else:
                    pos = _np.zeros(3)
                    pos[:2] = function(t)

                # Rotation at xi.
                if is_rot_funct:
                    rot = function_rotation(t)
                else:
                    r_prime = rp(t)
                    if is_3d_curve:
                        # Create the next triad via the smallest rotation mapping based
                        # on the last triad.
                        rot = _smallest_rotation(self.last_triad, r_prime)
                        self.last_triad = rot.copy()
                    else:
                        # The rotation simplifies in the 2d case.
                        rot = _Rotation([0, 0, 1], _np.arctan2(r_prime[1], r_prime[0]))

                # Set start values for the next iteration
                self.t_start_newton = t
                self.S_start_newton = S

                # Return the needed values for beam creation.
                return (pos, rot, S)

            return get_beam_position_and_rotation_at_xi

    # Now create the beam.
    # Get the length of the whole segment.
    length = S(interval[1])

    # Create the beam in the mesh
    created_sets = _create_beam_mesh_generic(
        mesh,
        beam_class=beam_class,
        material=material,
        function_generator=BeamFunctions(),
        interval=[0.0, length],
        interval_length=length,
        **kwargs,
    )

    if output_length:
        return (created_sets, length)
    else:
        return created_sets



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test general functionality of the 4C module with end-
to-end integration tests."""

import numpy as np
import pytest
import splinepy

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.function import Function
from beamme.core.geometry_set import GeometrySetNodes
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import (
    Beam3rHerm2Line3,
    get_four_c_reissner_beam,
)
from beamme.four_c.header_functions import (
    add_result_description,
    set_beam_to_solid_meshtying,
    set_header_static,
    set_runtime_output,
)
from beamme.four_c.input_file import InputFile
from beamme.four_c.model_importer import import_four_c_model
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.mesh_creation_functions.nurbs_generic import add_splinepy_nurbs_to_mesh


@pytest.mark.parametrize(
    ("coupling_type", "coupling_dof_type", "additional_identifier"),
    (
        (bme.bc.point_coupling, bme.coupling_dof.fix, "exact"),
        (
            bme.bc.point_coupling_penalty,
            {"POSITIONAL_PENALTY_PARAMETER": 10000, "ROTATIONAL_PENALTY_PARAMETER": 0},
            "penalty",
        ),
    ),
)
def test_integration_four_c_point_coupling(
    coupling_type,
    coupling_dof_type,
    additional_identifier,
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of point couplings for 4C."""

    # Create material and mesh
    material = get_default_test_beam_material(
        material_type="reissner", interaction_radius=2.0
    )
    mesh = Mesh()

    # Create a 2x2 grid of beams.
    for i in range(3):
        for j in range(2):
            create_beam_mesh_line(
                mesh, Beam3rHerm2Line3, material, [j, i, 0.0], [j + 1, i, 0.0]
            )
            create_beam_mesh_line(
                mesh, Beam3rHerm2Line3, material, [i, j, 0.0], [i, j + 1, 0.0]
            )

    # Couple the beams.
    mesh.couple_nodes(
        reuse_matching_nodes=True,
        coupling_type=coupling_type,
        coupling_dof_type=coupling_dof_type,
    )

    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        mesh,
    )


def test_integration_four_c_point_coupling_indirect(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that indirect point coupling works as expected."""

    material = get_default_test_beam_material(material_type="reissner")
    mesh = Mesh()

    beam_set_1 = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, material, [-1, 0, 0], [1, 0, 0]
    )
    beam_set_2 = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, material, [0, -1, 0.1], [0, 1, 0.1], n_el=3
    )
    for i_set, beam_set in enumerate([beam_set_1, beam_set_2]):
        data = {"COUPLING_ID": 3}
        if i_set == 0:
            data["PARAMETERS"] = {
                "POSITIONAL_PENALTY_PARAMETER": 1.1,
                "ROTATIONAL_PENALTY_PARAMETER": 1.2,
                "PROJECTION_VALID_FACTOR": 1.3,
            }
        mesh.add(
            BoundaryCondition(
                beam_set["line"],
                data=data,
                bc_type=bme.bc.point_coupling_indirect,
            )
        )

    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_four_c_fluid_element_section(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Add beam elements to an input file containing fluid elements."""

    input_file, _ = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            additional_identifier="import"
        )
    )

    beam_mesh = Mesh()
    material = get_default_test_beam_material(material_type="reissner")
    beam_mesh.add(material)

    create_beam_mesh_line(
        beam_mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        material,
        [0, -0.5, 0],
        [0, 0.2, 0],
        n_el=5,
    )

    input_file.add(beam_mesh)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


def test_integration_four_c_surface_to_surface_contact_import(
    assert_results_close, get_corresponding_reference_file_path
):
    """Test that surface-to-surface contact problems can be imported as
    expected."""

    input_file, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            additional_identifier="solid_mesh"
        ),
        convert_input_to_mesh=True,
    )

    input_file.add(mesh)

    # Compare with the reference file.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


def test_integration_four_c_nurbs_import(
    get_default_test_beam_material,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
    tmp_path,
):
    """Test if the import of a NURBS mesh works as expected.

    We first create the NURBS structure, dump it to disk and then read it again.
    Then we add the beam and beam-to-volume stuff.

    This script generates the 4C test case:
    beam3r_herm2line3_static_beam_to_solid_volume_meshtying_nurbs27_mortar_penalty_line4

    TODO: This test case basically covers that we can import NURBS from
    existing import files. However, this test case also contains a fully
    working 4C test case. So this case should be moved to the 4C tests and
    we should add a very basic unit test case if we can import NURBS.
    """

    # Create a third of the NURBS hollow cylinder
    base = splinepy.helpme.create.disk(
        outer_radius=0.3, inner_radius=0.2, angle=120, n_knot_spans=1
    )
    extruded = base.create.extruded(extrusion_vector=[0, 0, 1])
    extruded.elevate_degrees([0, 2])

    # Add NURBS to mesh (3 times to get the full cylinder)
    mesh = Mesh()
    mat = get_default_test_solid_material(material_type="st_venant_kirchhoff")
    element_description = {"KINEM": "nonlinear"}

    volume_set = GeometrySetNodes(geometry_type=bme.geo.volume)
    fix_set = GeometrySetNodes(geometry_type=bme.geo.surface)
    for i in range(3):
        patch_set = add_splinepy_nurbs_to_mesh(
            mesh, extruded, material=mat, data=element_description
        )
        volume_set = volume_set + patch_set["vol"]
        fix_set = fix_set + patch_set["surf_w_min"]
        mesh.add(patch_set)
        mesh.rotate(Rotation([0, 0, 1], np.pi * 2.0 / 3.0))

    mesh.add(
        BoundaryCondition(
            fix_set,
            data={
                "NUMDOF": 3,
                "ONOFF": [1, 1, 1],
                "VAL": [0, 0, 0],
                "FUNCT": [0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    mesh.add(
        BoundaryCondition(
            volume_set,
            data={"COUPLING_ID": 1},
            bc_type=bme.bc.beam_to_solid_volume_meshtying,
        )
    )

    # We need this because we get "double" CP from the splinepy object.
    mesh.couple_nodes(reuse_matching_nodes=True)
    # We need this to math the "old" result description
    mesh.rotate(Rotation([0, 0, 1], np.pi * 0.5))
    nurbs_input_file = InputFile()
    nurbs_input_file.add(mesh)
    nurbs_path = tmp_path / "nurbs_mesh.4C.yaml"
    nurbs_input_file.dump(
        nurbs_path,
        validate=False,
        add_header_default=False,
        add_header_information=False,
        add_footer_application_script=False,
    )

    # Create mesh and load solid file.
    input_file, mesh = import_four_c_model(input_file_path=nurbs_path)

    set_header_static(
        input_file,
        time_step=0.5,
        n_steps=2,
        tol_residuum=1e-14,
        tol_increment=1e-8,
    )
    set_beam_to_solid_meshtying(
        input_file,
        bme.bc.beam_to_solid_volume_meshtying,
        contact_discretization="mortar",
        mortar_shape="line4",
        penalty_parameter=1000,
        n_gauss_points=6,
        segmentation=True,
        binning_parameters={
            "binning_bounding_box": [-3, -3, -1, 3, 3, 5],
            "binning_cutoff_radius": 1,
        },
    )
    set_runtime_output(input_file)
    input_file["PROBLEM TYPE"]["SHAPEFCT"] = "Nurbs"
    input_file["IO"]["OUTPUT_BIN"] = True
    input_file["IO"]["STRUCT_DISP"] = True
    input_file["IO"]["VERBOSITY"] = "Standard"

    fun = Function([{"SYMBOLIC_FUNCTION_OF_TIME": "t"}])
    mesh.add(fun)

    # Create the beam material.
    material = get_default_test_beam_material(material_type="reissner")

    # Create the beams.
    set_1 = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, material, [0, 0, 0.95], [1, 0, 0.95], n_el=2
    )
    set_2 = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        material,
        [-0.25, -0.3, 0.85],
        [-0.25, 0.5, 0.85],
        n_el=2,
    )

    # Add boundary conditions on the beams.
    mesh.add(
        BoundaryCondition(
            set_1["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [0, 0, 0, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    mesh.add(
        BoundaryCondition(
            set_1["end"],
            {
                "NUMDOF": 9,
                "ONOFF": [0, 1, 0, 0, 0, 0, 0, 0, 0],
                "VAL": [0, 0.02, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, fun, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.neumann,
        )
    )
    mesh.add(
        BoundaryCondition(
            set_2["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [0, 0, 0, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    mesh.add(
        BoundaryCondition(
            set_2["end"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 0, 0, 0, 0, 0, 0, 0, 0],
                "VAL": [-0.06, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [fun, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.neumann,
        )
    )
    mesh.add(
        BoundaryCondition(
            set_1["line"] + set_2["line"],
            {"COUPLING_ID": 1},
            bc_type=bme.bc.beam_to_solid_volume_meshtying,
        )
    )

    # Add result checks.
    displacements = [
        [
            -5.14451531199392575e-01,
            -1.05846397823837826e-01,
            -1.77822866488512921e-01,
        ]
    ]
    nodes = [64]
    add_result_description(input_file, displacements, nodes)

    # Add the mesh to the input file
    input_file.add(mesh)

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


def test_integration_four_c_user_defined_boundary_condition(
    get_bc_data,
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Check if a user-defined boundary condition can be added."""

    mesh = Mesh()

    mat = get_default_test_beam_material(material_type="reissner")
    sets = create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 2, 3])
    mesh.add(
        BoundaryCondition(
            sets["line"], get_bc_data(), bc_type="DESIGN VOL ALE DIRICH CONDITIONS"
        )
    )

    # Compare the output of the mesh.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


@pytest.mark.parametrize("reuse_nodes", [False, True])
def test_integration_four_c_check_multiple_node_penalty_coupling(
    get_default_test_beam_material,
    reuse_nodes,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """For point penalty coupling constraints, we add multiple coupling
    conditions.

    This is checked in this test case. The flag reuse_nodes decides
    whether equal nodes are unified to a single node.
    """

    # Create mesh object
    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")
    mesh.add(mat)

    # Add three beams that have one common point
    create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0])
    create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [1, 0, 0], [2, 0, 0])
    create_beam_mesh_line(mesh, Beam3rHerm2Line3, mat, [1, 0, 0], [2, -1, 0])

    mesh.couple_nodes(
        reuse_matching_nodes=reuse_nodes,
        coupling_type=bme.bc.point_coupling_penalty,
        coupling_dof_type={
            "POSITIONAL_PENALTY_PARAMETER": 10000,
            "ROTATIONAL_PENALTY_PARAMETER": 0,
        },
    )
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier="reuse" if reuse_nodes else ""
        ),
        mesh,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the header functions for 4C."""

import pytest

from beamme.core.conf import bme
from beamme.four_c.header_functions import (
    set_beam_contact_runtime_output,
    set_beam_contact_section,
    set_beam_to_solid_meshtying,
    set_binning_strategy_section,
    set_header_static,
    set_runtime_output,
)
from beamme.four_c.input_file import InputFile


def test_integration_four_c_header_functions_static(
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test the default static header function."""

    # Create input file.
    input_file = InputFile()

    # Set the header.
    set_header_static(input_file, time_step=0.1, n_steps=17, load_lin=True)
    set_runtime_output(input_file, output_triad=False)

    set_binning_strategy_section(
        input_file, binning_bounding_box=[1, 2, 3, 4, 5, 6], binning_cutoff_radius=0.69
    )

    set_beam_to_solid_meshtying(
        input_file,
        bme.bc.beam_to_solid_volume_meshtying,
        contact_discretization="mortar",
    )

    set_beam_to_solid_meshtying(
        input_file,
        bme.bc.beam_to_solid_surface_meshtying,
        contact_discretization="gp",
        coupling_type="consistent_fad",
        segmentation=False,
    )

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


@pytest.mark.parametrize(
    "additional_identifier, time_step, n_steps, total_time",
    [
        ("all", 0.01, 17, 2.1),
        ("no_total_time", 0.01, 17, None),
        ("no_n_steps", 0.01, None, 2.1),
        ("no_time_step", None, 17, 2.1),
    ],
)
def test_integration_four_c_header_functions_static_time(
    get_corresponding_reference_file_path,
    assert_results_close,
    additional_identifier,
    time_step,
    n_steps,
    total_time,
):
    """Test the time setting options in the static header functions."""

    input_file = InputFile()

    set_header_static(
        input_file, time_step=time_step, n_steps=n_steps, total_time=total_time
    )

    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        input_file,
    )


def test_integration_four_c_header_functions_static_prestress(
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test the static header function with non default prestressing
    parameter."""

    # Create input file.
    input_file = InputFile()

    # Set the header.
    set_header_static(
        input_file,
        time_step=0.1,
        n_steps=17,
        load_lin=True,
        prestress="mulf",
        prestress_time=1,
    )
    set_beam_to_solid_meshtying(
        input_file,
        bme.bc.beam_to_solid_volume_meshtying,
        contact_discretization="mortar",
        binning_parameters={
            "binning_bounding_box": [1, 2, 3, 4, 5, 6],
            "binning_cutoff_radius": 0.69,
        },
        couple_restart=True,
    )

    set_beam_to_solid_meshtying(
        input_file,
        bme.bc.beam_to_solid_surface_meshtying,
        contact_discretization="gp",
        segmentation=False,
        couple_restart=False,
    )

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


def test_integration_four_c_header_functions_stress_output(
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test the static header function with non default stress output
    parameter."""

    # Create input file.
    input_file = InputFile()

    # Set the header.
    set_header_static(
        input_file,
        time_step=0.1,
        n_steps=17,
        load_lin=True,
        write_stress="cauchy",
        write_strain="gl",
    )
    set_runtime_output(input_file, output_stress_strain=True)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


def test_integration_four_c_header_functions_beam_interaction(
    get_corresponding_reference_file_path, assert_results_close
):
    """Test the beam-to-beam contact header function with default parameter."""

    # Create input file.
    input_file = InputFile()

    # Add Beam contact section to file.
    set_beam_contact_section(
        input_file,
        binning_parameters={
            "binning_cutoff_radius": 5,
            "binning_bounding_box": [-1, -2, -3, 1, 2, 3],
        },
        beam_interaction_parameters={"repartition_strategy": "adaptive"},
    )

    # Add per default the runtime output.
    set_beam_contact_runtime_output(input_file)

    # Compare the output.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


@pytest.mark.parametrize(
    ("create_nox_file", "nox_xml_file_kwarg", "xml_relative_path"),
    [
        (False, None, None),
        (True, None, "nonlinear_solver_parameters.nox.xml"),
        (True, "custom_name.xml", "custom_name.xml"),
    ],
)
def test_integration_four_c_header_functions_nonlinear_solver_parameters(
    get_corresponding_reference_file_path,
    assert_results_close,
    create_nox_file,
    nox_xml_file_kwarg,
    xml_relative_path,
    tmp_path,
):
    """Test that the NOX xml is exported correctly."""

    input_file = InputFile()
    set_header_static(
        input_file,
        total_time=1.0,
        n_steps=1,
        tol_increment=1e-4,
        tol_residuum=1e-5,
        max_iter=25,
        create_nox_file=create_nox_file,
    )
    input_file.dump(
        tmp_path / "nonlinear_solver_parameters.4C.yaml",
        nox_xml_file=nox_xml_file_kwarg,
        add_header_default=False,
        add_header_information=False,
        add_footer_application_script=False,
    )

    if create_nox_file:
        # Check the xml path in the input file
        assert input_file["STRUCT NOX/Status Test"]["XML File"] == xml_relative_path

        # Check the created xml
        assert_results_close(
            get_corresponding_reference_file_path(extension="xml"),
            tmp_path / xml_relative_path,
        )

    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=("with_nox" if create_nox_file else "")
            + ("_custom_name" if nox_xml_file_kwarg else "")
        ),
        tmp_path / "nonlinear_solver_parameters.4C.yaml",
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This function converts the DBC monitor log files to Neumann boundary
conditions in a mesh."""

from typing import Optional as _Optional

import numpy as _np
import yaml as _yaml

from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.conf import bme as _bme
from beamme.core.function import Function as _Function
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.mesh import Mesh as _Mesh
from beamme.four_c.function_utility import (
    create_linear_interpolation_function as _create_linear_interpolation_function,
)
from beamme.four_c.function_utility import (
    ensure_length_of_function_array as _ensure_length_of_function_array,
)


def linear_time_transformation(
    time, values, time_span, *, flip=False, valid_start_and_end_point=False
):
    """Performs a transformation of the time to a new interval with an
    appropriate value vector.

    Args
    ----
    time: _np.array
        array with time values
    values: _np.array
        corresponding values to time
    time_span: [list] with 2 or 3 entries:
        time_span[0:2] defines the time interval to which the initial time interval should be scaled.
        time_span[3] optional timepoint to repeat last value
    flip: Bool
        Flag if the values should be reversed
    valid_start_and_end_point: Bool
        optionally adds a valid starting point at t=0 and timespan[3] if provided
    """

    # flip values if desired and adjust time
    if flip is True:
        values = _np.flipud(values)
        time = _np.flip(-time) + time[-1]

    # transform time to interval
    min_t = _np.min(time)
    max_t = _np.max(time)

    # scaling/transforming the time into the user defined time
    time = time_span[0] + (time - min_t) * (time_span[1] - time_span[0]) / (
        max_t - min_t
    )

    # ensure that start time is valid
    if valid_start_and_end_point and time[0] > 0.0:
        # add starting time 0
        time = _np.append(0.0, time)

        # add first coordinate again at the beginning of the array
        if len(values.shape) == 1:
            values = _np.append(values[0], values)
        else:
            values = _np.append(values[0], values).reshape(
                values.shape[0] + 1, values.shape[1]
            )

    # repeat last value at provided time point
    if valid_start_and_end_point and len(time_span) > 2:
        if time_span[2] > time_span[1]:
            time = _np.append(time, time_span[2])
            values = _np.append(values, values[-1]).reshape(
                values.shape[0] + 1, values.shape[1]
            )
    if not valid_start_and_end_point and len(time_span) > 2:
        raise Warning("You specified unnecessarily a third component of time_span.")

    return time, values


def read_dbc_monitor_file(file_path):
    """Load the Dirichlet boundary condition monitor log and return the data as
    well as the nodes of this boundary condition.

    Args
    ----
    file_path: str
        Path to the Dirichlet boundary condition monitor log.

    Return
    ----
    [node_ids], [time], [force], [moment]
    """

    with open(file_path, "r") as f:
        dbc_monitor_file = _yaml.safe_load(f)

    nodes = dbc_monitor_file["dbc monitor condition"]["node gids"]

    time = []
    force = []
    moment = []
    for time_step_data in dbc_monitor_file["dbc monitor condition data"]:
        time.append(time_step_data["time"])
        force.append(time_step_data["f"])
        moment.append(time_step_data["m"])

    return (
        nodes,
        _np.asarray(time, dtype=float),
        _np.asarray(force, dtype=float),
        _np.asarray(moment, dtype=float),
    )


def add_point_neuman_condition_to_mesh(
    mesh: _Mesh,
    nodes: list[int],
    function_array: list[_Function],
    force: _np.ndarray,
    *,
    n_dof: int = 3,
):
    """Adds a Neumann boundary condition to a mesh for the given node_ids with
    the function_array and force values by creating a new geometry set.

    Args
    ----
    mesh: Mesh
        Mesh where the boundary conditions are added to
    nodes: [node_id]
        list containing the ids of the nodes for the condition
    function_array: [function]
        list with functions
    force: [_np.ndarray]
        values to scale the function array with
    n_dof: int
        Number of DOFs per node.
    """

    # check if the dimensions of force and functions match
    if force.size != 3:
        raise ValueError(
            f"The forces vector must have dimensions [3x1] not [{force.size}x1]"
        )

    function_array = _ensure_length_of_function_array(function_array, 3)

    # Add the function to the mesh, if they are not previously added.
    for function in function_array:
        mesh.add(function)

    # Create GeometrySet with nodes.
    mesh_nodes = [mesh.nodes[i_node] for i_node in nodes]
    geo = _GeometrySet(mesh_nodes)

    # Create the Boundary Condition.
    bc = _BoundaryCondition(
        geo,
        {
            "NUMDOF": n_dof,
            "ONOFF": [1, 1, 1] + [0] * (n_dof - 3),
            "VAL": force.tolist() + [0] * (n_dof - 3),
            "FUNCT": function_array + [0] * (n_dof - 3),
        },
        bc_type=_bme.bc.neumann,
    )
    mesh.add(bc)


def dbc_monitor_to_mesh_all_values(
    mesh: _Mesh,
    file_path: str,
    *,
    steps: list[int] = [],
    time_span: list[int] = [0, 1, 2],
    type: _Optional[str] = "linear",
    flip_time_values: bool = False,
    functions: list[_Function] = [],
    **kwargs,
):
    """Extracts all the force values of the monitored Dirichlet boundary
    condition and converts them into a Function with a Neumann boundary
    condition for a given mesh. The monitor log force values must be obtained
    from a previous simulation with constant step size. The discretization of
    the previous simulation must be identical to the one within the mesh. The
    extracted force values are passed to a linear interpolation 4C-function. It
    is advisable to only call this function once all nodes have been added to
    the mesh.

    Args
    ----
    mesh: Mesh
        The mesh where the created Neumann boundary condition is added
        to. The nodes(e.g., discretization) referred to in the log file must match with the ones
        in the mesh.
    file_path: str
        Path to the Dirichlet boundary condition log file.
    steps: [int,int]
        Index range of which the force values are extracted. Default 0 and -1 extracts every point from the array.
    time_span: [t1, t2, t3] in float
        Transforms the given time array into this specific format.
        The time array always starts at 0 and ends at t3 to ensure a valid simulation.
    type: str or linear
        two types are available:
            1) "linear": not specified simply extract all values and apply them between time interval t1 and t2.
            2) "hat": puts the values first until last value is reached and then decays them back to first value.
            Interpolation starting from t1 going to the last value at (t1+t2)/2 and going back to the value at time t2.
    flip_time_values: bool
        indicates, if the extracted forces should be flipped or rearranged wrt. to the time
        For flip_time_values=true, the forces at the final time are applied at t_start.
    functions: [Function, Function, Function]
        Array consisting of 3 custom functions(x,y,z). The value for boundary condition is selected from the last steps.
    """

    nodes, time, force, _ = read_dbc_monitor_file(file_path)

    # The forces are the negative reactions at the Dirichlet boundaries.
    force *= -1.0

    # if special index range is provided use it
    if steps:
        time = time[steps[0] : steps[1] + 1]
        force = force[steps[0] : steps[1] + 1, :]
    else:
        # otherwise define steps from start to end
        steps = [0, -1]

    # apply transformations to time and forces according to the schema
    if type == "linear":
        if not len(time_span) == 2:
            raise ValueError(
                f"Please provide a time_span with size 1x2 not {len(time_span)}"
            )

        time, force = linear_time_transformation(
            time, force, time_span, flip=flip_time_values
        )
        if len(functions) != 3:
            print("Please provide a list with three valid Functions.")

    elif type == "hat":
        if not len(time_span) == 3:
            raise ValueError(
                f"Please provide a time_span with size 1x3 not {len(time_span)}"
            )

        if functions:
            print(
                "You selected type",
                type,
                ", however the provided functions ",
                functions,
                " are overwritten.",
            )
        functions = []

        # create the two intervals
        time1, force1 = linear_time_transformation(
            time, force, time_span[0:2], flip=flip_time_values
        )
        time2, force2 = linear_time_transformation(
            time, force, time_span[1:3], flip=(not flip_time_values)
        )

        # remove first element since it is duplicated zero
        _np.delete(time2, 0)
        _np.delete(force2, 0)

        # add the respective force
        time = _np.concatenate((time1, time2[1:]))
        force = _np.concatenate((force1, force2[1:]), axis=0)

    else:
        raise ValueError(
            "The selected type: "
            + str(type)
            + " is currently not supported. Feel free to add it here."
        )

    # overwrite the function, if one is provided since for the specific types the function is generated
    if not type == "linear":
        for dim in range(force.shape[1]):
            # create a linear function with the force values per dimension
            fun = _create_linear_interpolation_function(
                time, force[:, dim], function_type="SYMBOLIC_FUNCTION_OF_TIME"
            )

            # add the function to the mesh
            mesh.add(fun)

            # store function
            functions.append(fun)

        # now set forces to 1 since the force values are already extracted in the function's values
        force = _np.zeros_like(force) + 1.0

    elif len(functions) != 3:
        raise ValueError("Please provide functions with ")

    # Create condition in mesh
    add_point_neuman_condition_to_mesh(
        mesh, nodes, functions, force[steps[1]], **kwargs
    )


def dbc_monitor_to_mesh(
    mesh: _Mesh,
    file_path: str,
    *,
    step: int = -1,
    function: _Function,
    **kwargs,
):
    """Converts the last value of a Dirichlet boundary condition monitor log to
    a Neumann boundary condition in the mesh.

    Args
    ----
    mesh: Mesh
        The mesh where the created Neumann boundary condition is added to.
        The nodes referred to in the log file have to match with the ones
        in the mesh. It is advisable to only call this function once
        all nodes have been added to the mesh.
    file_path: str
        Path to the Dirichlet boundary condition log file.
    step: int
        Step values to be used. Default is -1, i.e. the last step.
    function: Function
        Function for the Neumann boundary condition.
    """

    # read the force
    nodes, _, force, _ = read_dbc_monitor_file(file_path)

    # The forces are the negative reactions at the Dirichlet boundaries.
    force *= -1.0

    # Create condition in mesh
    add_point_neuman_condition_to_mesh(
        mesh, nodes, [function] * 3, force[step], **kwargs
    )
